#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "substdio.h"
#include "subfd.h"
#include "stralloc.h"
#include "readwrite.h"
#include "error.h"
#include "fd.h"
#include "sig.h"
#include "wait.h"
#include "ip.h"
#include "ipalloc.h"
#include "dns.h"
#include "str.h"
#include "case.h"
#include "byte.h"
#include "sgetopt.h"
#include "remoteinfo.h"
#include "exit.h"
#include "scan.h"
#include "fmt.h"
#include "env.h"

int verbosity = 1;
char temp[IPFMT + FMT_ULONG];

void out(s) char *s; { if (verbosity) substdio_puts(subfderr,s); }
void flush() { if (verbosity) substdio_flush(subfderr); }

void usage()
{ out("tcpserver: usage: \
tcpserver [ -qQvdDoOpPhHrR1 ] \
[ -bbacklog ] [ -climit ] [ -ttimeout ] [ -llocalname ] [ -ggid ] [ -uuid ] \
host port program\n"); flush(); _exit(100); }
void die(s) char *s;
{ out("tcpserver: fatal: "); out(s); out("\n"); flush(); _exit(111); }
void diep(s) char *s;
{ char *x = error_str(errno);
  out("tcpserver: fatal: "); out(s); out(": ");
  out(x); out("\n"); flush(); _exit(111); }
void nomem() { die("out of memory"); }
void die_gid() { die("unable to change gid"); }
void die_uid() { die("unable to change uid"); }
void warnfork() { char *x = error_str(errno);
  out("tcpserver: warning: unable to fork: "); out(x); out("\n"); flush(); }

unsigned long numchildren = 0;

void infoconnected(pid,ipr) unsigned long pid; struct ip_address *ipr;
{
  if (verbosity < 2) return;
  out("tcpserver: pid ");
  temp[fmt_ulong(temp,pid)] = 0;
  out(temp);
  out(" num ");
  temp[fmt_ulong(temp,numchildren)] = 0;
  out(temp);
  out(" from ");
  temp[ip_fmt(temp,ipr)] = 0;
  out(temp);
  out("\n");
  flush();
}

char printbuf[16];
struct substdio print = SUBSTDIO_FDBUF(write,1,printbuf,sizeof(printbuf));
void printport(port) unsigned long port;
{
  unsigned long len;
  len = fmt_ulong(temp,port);
  temp[len++] = '\n';
  substdio_putflush(&print,temp,len);
}

void sigterm() { _exit(0); }

void sigchld()
{
  int wstat;
  int pid;
 
  while ((pid = wait_nohang(&wstat)) > 0) {
    if (numchildren) --numchildren;
    if (verbosity >= 2) {
      out("tcpserver: end ");
      temp[fmt_ulong(temp,(unsigned long) pid)] = 0;
      out(temp);
      out(" status ");
      temp[fmt_ulong(temp,(unsigned long) wstat)] = 0;
      out(temp);
      out("\n");
      flush();
    }
  }
}

int flagkillopts = 1;
int flagdelay = 1;
int flagremoteinfo = 1;
int flagremotehost = 1;
int flagparanoid = 0;
int flag1 = 0;
unsigned long backlog = 20;
unsigned long timeout = 26;
unsigned long uid = 0;
unsigned long gid = 0;
char *forcelocal = 0;
unsigned long limit = 40;

stralloc tmp = {0};
ipalloc ia = {0};

struct sockaddr_in salocal;
struct ip_address iplocal;
unsigned long portlocal;
struct sockaddr_in saremote;
struct ip_address ipremote;
unsigned long portremote;

void main(argc,argv)
int argc;
char **argv;
{
  int s;
  int t;
  int dummy;
  int opt;
  char *hostname;
  char *portname;
  struct servent *se;
  int j;
 
  while ((opt = getopt(argc,argv,"dDvqQhHrR1t:u:g:l:b:c:pPoO")) != opteof)
    switch(opt) {
      case 'b': scan_ulong(optarg,&backlog); break;
      case 'c': scan_ulong(optarg,&limit); break;
      case 'd': flagdelay = 1; break;
      case 'D': flagdelay = 0; break;
      case 'v': verbosity = 2; break;
      case 'q': verbosity = 0; break;
      case 'Q': verbosity = 1; break;
      case 'P': flagparanoid = 0; break;
      case 'p': flagparanoid = 1; break;
      case 'O': flagkillopts = 1; break;
      case 'o': flagkillopts = 0; break;
      case 'H': flagremotehost = 0; break;
      case 'h': flagremotehost = 1; break;
      case 'R': flagremoteinfo = 0; break;
      case 'r': flagremoteinfo = 1; break;
      case 't': scan_ulong(optarg,&timeout); break;
      case 'g': scan_ulong(optarg,&gid); break;
      case 'u': scan_ulong(optarg,&uid); break;
      case '1': flag1 = 1; break;
      case 'l': forcelocal = optarg; break;
      default: usage();
    }
  argc -= optind;
  argv += optind;
 
  hostname = *argv++;
  if (!hostname) usage();
  portname = *argv++;
  if (!portname) usage();
  if (!*argv) usage();
 
  sig_pipeignore();
  sig_termcatch(sigterm);
  sig_childcatch(sigchld);
 
  dns_init(1);
 
  byte_zero(&salocal,sizeof(salocal));
  salocal.sin_family = AF_INET;
 
  if (!portname[scan_ulong(portname,&portlocal)])
    salocal.sin_port = htons((unsigned short) portlocal);
  else {
    se = getservbyname(portname,"tcp");
    if (!se) die("unable to figure out port number");
    salocal.sin_port = se->s_port;
  }
 
  if (!str_diff(hostname,"0")) hostname = "0.0.0.0";
 
  if (hostname[ip_scan(hostname,&iplocal)]) {
    if (!stralloc_copys(&tmp,hostname)) nomem();
    switch(dns_ip(&ia,&tmp)) {
      case DNS_MEM: nomem();
      case DNS_HARD: die("unable to figure out host address");
      case DNS_SOFT: die("temporarily unable to figure out host address");
    }
    if (!ia.len) die("no IP addresses for that host");
    byte_copy(&iplocal,4,&ia.ix[0].ip);
  }
 
  byte_copy(&salocal.sin_addr,4,&iplocal);
  s = socket(AF_INET,SOCK_STREAM,0);
  if (s == -1) diep("unable to create socket");
 
  {
    int opt = 1;
    setsockopt(s,SOL_SOCKET,SO_REUSEADDR,&opt,sizeof(opt));
  }

  if (bind(s,(struct sockaddr *) &salocal,sizeof(salocal)) == -1)
    diep("unable to bind");
  dummy = sizeof(salocal);
  if (getsockname(s,(struct sockaddr *) &salocal,&dummy) == -1)
    diep("unable to get local address");
  if (listen(s,backlog) == -1)
    diep("unable to listen");
 
  if (gid) if (setgid(gid) == -1) die_gid();
  if (uid) if (setuid(uid) == -1) die_uid();
 
  if (!env_init()) nomem();
  if (!env_put("PROTO=TCP")) nomem();
  if (!env_unset("TCPLOCALHOST")) nomem();
  if (!env_unset("TCPREMOTEHOST")) nomem();
  if (!env_unset("TCPREMOTEINFO")) nomem();
 
  if (forcelocal)
    if (!env_put2("TCPLOCALHOST",forcelocal)) nomem();
 
  portlocal = ntohs(salocal.sin_port);
  temp[fmt_ulong(temp,portlocal)] = 0;
  if (!env_put2("TCPLOCALPORT",temp)) nomem();
  if (flag1) printport(portlocal);
 
  close(0);
  close(1);
  sig_childblock();
 
  for (;;) {
    while (numchildren >= limit) sig_pause();
    sig_childunblock();
 
    dummy = sizeof(saremote);
    t = accept(s,(struct sockaddr *) &saremote,&dummy);
    if (t == -1) continue;
    portremote = ntohs(saremote.sin_port);
    byte_copy(&ipremote,4,&saremote.sin_addr);
 
    sig_childblock();
 
    switch(fork()) {
      default:
        ++numchildren;
        break;
      case -1:
        warnfork();
        break;
      case 0:
        infoconnected((unsigned long) getpid(),&ipremote);
        close(s);
        if (flagkillopts) {
          setsockopt(t,IPPROTO_IP,1,(char *) 0,0); /* 1 == IP_OPTIONS */
          /* if it fails, bummer */
        }
        if (!flagdelay) {
          int opt = 1;
          setsockopt(t,IPPROTO_TCP,1,&opt,sizeof(opt)); /* 1 == TCP_NODELAY */
          /* if it fails, bummer */
        }
        
        dummy = sizeof(salocal);
        if (getsockname(t,(struct sockaddr *) &salocal,&dummy) == -1)
          diep("unable to get local address");
        byte_copy(&iplocal,4,&salocal.sin_addr);
       
        temp[ip_fmt(temp,&iplocal)] = 0;
        if (!env_put2("TCPLOCALIP",temp)) nomem();
        if (!forcelocal)
          switch(dns_ptr(&tmp,&iplocal)) {
            case DNS_MEM: nomem();
            case 0:
              if (!stralloc_0(&tmp)) nomem();
              case_lowers(tmp.s);
              if (!env_put2("TCPLOCALHOST",tmp.s)) nomem();
          }
 
        temp[fmt_ulong(temp,portremote)] = 0;
        if (!env_put2("TCPREMOTEPORT",temp)) nomem();
        temp[ip_fmt(temp,&ipremote)] = 0;
        if (!env_put2("TCPREMOTEIP",temp)) nomem();
        if (flagremotehost)
          switch(dns_ptr(&tmp,&ipremote)) {
            case DNS_MEM: nomem();
            case 0:
              if (flagparanoid) {
                if (dns_ip(&ia,&tmp) != 0) break;
                for (j = 0;j < ia.len;++j)
                  if (!byte_diff(&ipremote,4,&ia.ix[j].ip))
                    break;
                if (j == ia.len)
                  break;
              }
              if (!stralloc_0(&tmp)) nomem();
              case_lowers(tmp.s);
              if (!env_put2("TCPREMOTEHOST",tmp.s)) nomem();
          }
        if (flagremoteinfo) {
          char *rinfo;
          rinfo = remoteinfo_get(&ipremote,portremote,&iplocal,portlocal,(int) timeout);
          if (rinfo)
            if (!env_put2("TCPREMOTEINFO",rinfo)) nomem();
        }
        if (fd_move(0,t) == -1) diep("unable to set up descriptor 0");
        if (fd_copy(1,0) == -1) diep("unable to set up descriptor 1");
        sig_childunblock();
        sig_childdefault();
        sig_termdefault();
        sig_pipedefault();
        execvp(*argv,argv);
        diep("unable to execute");
    }
    close(t);
  }
}
